import re
from typing import Optional, Type, Iterable, Mapping, Any, Dict, NamedTuple

from pydantic import BaseModel

from flask_pydantic_spec.constants import OPENAPI_SCHEMA_TEMPLATE


class ResponseBase:
    """
    Base Class for Response Types
    """

    def has_model(self) -> bool:
        raise NotImplementedError

    def find_model(self, code: int) -> Optional[Type[BaseModel]]:
        raise NotImplementedError

    @property
    def models(self) -> Iterable[Type[BaseModel]]:
        raise NotImplementedError

    def generate_spec(self) -> Mapping[str, Any]:
        raise NotImplementedError


class ResponseModel(NamedTuple):
    model: Type[BaseModel]
    is_list: bool = False


class Response(ResponseBase):
    """
    response object

    :param args: list of HTTP status code, format('HTTP_[0-9]{3}'), 'HTTP200'
    :param kwargs: dict of <HTTP status code>: <`pydantic.BaseModel`> or None
    You can also pass `validate` into the kwargs to disable output validation of this model
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self.validate = True
        self.codes = []
        for item in args:
            assert item in DEFAULT_CODE_DESC, "invalid HTTP status code"
            self.codes.append(item)

        self.code_models: Dict[str, ResponseModel] = {}
        for key, value in kwargs.items():
            if key.lower() == "validate":
                assert isinstance(value, bool)
                self.validate = value
            else:
                assert key in DEFAULT_CODE_DESC, "invalid HTTP status code"
                if value:
                    if self.is_list_type(value):
                        assert issubclass(
                            value.__args__[0], BaseModel
                        ), "invalid `pydantic.BaseModel`"
                        self.code_models[key] = ResponseModel(value.__args__[0], True)
                    else:
                        assert issubclass(value, BaseModel), "invalid `pydantic.BaseModel`"
                        self.code_models[key] = ResponseModel(value, False)
                else:
                    self.codes.append(key)

    @staticmethod
    def is_list_type(value: Any) -> bool:
        return hasattr(value, "__origin__") and value.__origin__ is list

    def has_model(self) -> bool:
        """
        :returns: boolean -- does this response has models or not
        """
        return True if self.code_models else False

    def find_model(self, code: int) -> Optional[Type[BaseModel]]:
        """
        :param code: ``r'\\d{3}'``
        """
        response_model = self.code_models.get(f"HTTP_{code}")
        if response_model:
            return response_model.model
        return None

    @property
    def models(self) -> Iterable[Type[BaseModel]]:
        """
        :returns:  dict_values -- all the models in this response
        """
        return [i.model for i in self.code_models.values()]

    def generate_spec(self) -> Dict[str, Any]:
        """
        generate the spec for responses

        :returns: JSON
        """
        responses: Dict[str, Any] = {}
        for code in self.codes:
            response_code = _parse_code(code)
            if response_code:
                responses[response_code] = {"description": DEFAULT_CODE_DESC[code]}

        for code, response_model in self.code_models.items():
            response_code = _parse_code(code)
            if response_code:
                schema = self.get_schema(response_model.model, is_list=response_model.is_list)
                responses[response_code] = {
                    "description": DEFAULT_CODE_DESC[code],
                    "content": {"application/json": schema},
                }

        return responses

    @staticmethod
    def get_schema(model: Type[BaseModel], is_list: bool = False) -> Mapping[str, Any]:
        ref_schema = {"$ref": f"#/components/schemas/{model.__name__}"}
        if is_list:
            return {"schema": {"type": "array", "items": ref_schema}}
        return {"schema": ref_schema}


class FileResponse(ResponseBase):
    def __init__(self, content_type: str = "application/octet-stream"):
        self.content_type = content_type

    def has_model(self) -> bool:
        """
        File response cannot have a model
        """
        return False

    @property
    def models(self) -> Iterable[Type[BaseModel]]:
        return []

    def generate_spec(self) -> Mapping[str, Any]:
        responses = {
            "200": {
                "description": DEFAULT_CODE_DESC["HTTP_200"],
                "content": {self.content_type: {"schema": {"type": "string", "format": "binary"}}},
            },
            "404": {"description": DEFAULT_CODE_DESC["HTTP_404"]},
        }

        return responses


class RequestBase:
    def has_model(self) -> bool:
        raise NotImplementedError

    def generate_spec(self) -> Mapping[str, Any]:
        raise NotImplementedError


class Request(RequestBase):
    def __init__(
        self,
        model: Optional[Type[BaseModel]] = None,
        content_type: str = "application/json",
        encoding: str = "binary",
    ) -> None:
        self.content_type = content_type
        self.model = model
        self.encoding = encoding

    def has_model(self) -> bool:
        return self.model is not None

    def generate_spec(self) -> Mapping[str, Any]:
        if self.content_type == "application/octet-stream":
            return {
                "content": {
                    self.content_type: {"schema": {"type": "string", "format": self.encoding}}
                }
            }
        else:
            assert self.model is not None
            return {
                "content": {
                    self.content_type: {
                        "schema": {"$ref": f"#/components/schemas/{self.model.__name__}"}
                    }
                }
            }


class MultipartFormRequest(RequestBase):
    def __init__(
        self,
        model: Optional[Type[BaseModel]] = None,
        file_key: str = "file",
        encoding: str = "binary",
    ):
        self.content_type = "multipart/form-data"
        self.model = model
        self.file_key = file_key
        self.encoding = encoding

    def has_model(self) -> bool:
        return self.model is not None

    def generate_spec(self) -> Mapping[str, Any]:
        model_spec = self.model.schema(ref_template=OPENAPI_SCHEMA_TEMPLATE) if self.model else None
        if model_spec:
            additional_properties = model_spec["properties"]
        else:
            additional_properties = {}

        return {
            "content": {
                "multipart/form-data": {
                    "schema": {
                        "type": "object",
                        "properties": {
                            **additional_properties,
                            self.file_key: {
                                "type": "string",
                                "format": self.encoding,
                            },
                        },
                    }
                }
            }
        }


HTTP_CODE = re.compile(r"^HTTP_(?P<code>\d{3})$")


def _parse_code(http_code: str) -> Optional[str]:
    """
    get the code of this HTTP status

    :param str http_code: format like ``HTTP_200``
    """
    match = HTTP_CODE.match(http_code)
    if not match:
        return None
    return match.group("code")


# according to https://tools.ietf.org/html/rfc2616#section-10
# https://tools.ietf.org/html/rfc7231#section-6.1
# https://developer.mozilla.org/sv-SE/docs/Web/HTTP/Status
DEFAULT_CODE_DESC = {
    # Information 1xx
    "HTTP_100": "Continue",
    "HTTP_101": "Switching Protocols",
    # Successful 2xx
    "HTTP_200": "OK",
    "HTTP_201": "Created",
    "HTTP_202": "Accepted",
    "HTTP_203": "Non-Authoritative Information",
    "HTTP_204": "No Content",
    "HTTP_205": "Reset Content",
    "HTTP_206": "Partial Content",
    # Redirection 3xx
    "HTTP_300": "Multiple Choices",
    "HTTP_301": "Moved Permanently",
    "HTTP_302": "Found",
    "HTTP_303": "See Other",
    "HTTP_304": "Not Modified",
    "HTTP_305": "Use Proxy",
    "HTTP_306": "(Unused)",
    "HTTP_307": "Temporary Redirect",
    "HTTP_308": "Permanent Redirect",
    # Client Error 4xx
    "HTTP_400": "Bad Request",
    "HTTP_401": "Unauthorized",
    "HTTP_402": "Payment Required",
    "HTTP_403": "Forbidden",
    "HTTP_404": "Not Found",
    "HTTP_405": "Method Not Allowed",
    "HTTP_406": "Not Acceptable",
    "HTTP_407": "Proxy Authentication Required",
    "HTTP_408": "Request Timeout",
    "HTTP_409": "Conflict",
    "HTTP_410": "Gone",
    "HTTP_411": "Length Required",
    "HTTP_412": "Precondition Failed",
    "HTTP_413": "Request Entity Too Large",
    "HTTP_414": "Request-URI Too Long",
    "HTTP_415": "Unsupported Media Type",
    "HTTP_416": "Requested Range Not Satisfiable",
    "HTTP_417": "Expectation Failed",
    "HTTP_418": "I'm a teapot",
    "HTTP_421": "Misdirected Request",
    "HTTP_422": "Unprocessable Entity",
    "HTTP_423": "Locked",
    "HTTP_424": "Failed Dependency",
    "HTTP_425": "Too Early",
    "HTTP_426": "Upgrade Required",
    "HTTP_428": "Precondition Required",
    "HTTP_429": "Too Many Requests",
    "HTTP_431": "Request Header Fields Too Large",
    "HTTP_451": "Unavailable For Legal Reasons",
    # Server Error 5xx
    "HTTP_500": "Internal Server Error",
    "HTTP_501": "Not Implemented",
    "HTTP_502": "Bad Gateway",
    "HTTP_503": "Service Unavailable",
    "HTTP_504": "Gateway Timeout",
    "HTTP_505": "HTTP Version Not Supported",
    "HTTP_506": "Variant Also negotiates",
    "HTTP_507": "Insufficient Sotrage",
    "HTTP_508": "Loop Detected",
    "HTTP_511": "Network Authentication Required",
}



from enum import Enum
import gzip
import json
import logging

from typing import Optional, Mapping, Callable, Any, Tuple, List, Type, Iterable, Dict
from dataclasses import dataclass

from pydantic import ValidationError, BaseModel
from flask import (
    request,
    abort,
    make_response,
    jsonify,
    Request as FlaskRequest,
    Flask,
    Response as FlaskResponse,
)
from werkzeug.datastructures import Headers
from werkzeug.routing import Rule, parse_converter_args

from .config import Config
from .page import PAGES
from .types import ResponseBase, RequestBase
from .utils import parse_multi_dict, parse_rule


@dataclass
class Context:
    query: Optional[BaseModel]
    body: Optional[BaseModel]
    headers: Optional[BaseModel]
    cookies: Optional[BaseModel]


class FlaskBackend:
    def __init__(self, validator: Any) -> None:
        self.validator = validator
        self.config: Config = validator.config
        self.logger: logging.Logger = logging.getLogger(__name__)

    def find_routes(self) -> Any:
        for rule in self.app.url_map.iter_rules():
            if any(str(rule).startswith(path) for path in (f"/{self.config.PATH}", "/static")):
                continue
            yield rule

    def bypass(self, func: Callable, method: str) -> bool:
        if method in ["HEAD", "OPTIONS"]:
            return True
        return False

    def parse_func(self, route: Any) -> Any:
        func = self.app.view_functions[route.endpoint]
        for method in route.methods:
            yield method, func

    def parse_path(self, route: Rule) -> Tuple[str, List[Any]]:
        subs = []
        parameters = []

        for converter, arguments, variable in parse_rule(route):
            if converter is None:
                subs.append(variable)
                continue
            subs.append(f"{{{variable}}}")

            args: Iterable[Any] = []
            kwargs: Dict[str, Any] = {}

            if arguments:
                args, kwargs = parse_converter_args(arguments)

            schema = None
            # See: https://werkzeug.palletsprojects.com/en/2.3.x/routing/#built-in-converters
            if converter == "any":
                schema = {
                    "type": "string",
                    "enum": list(args),
                }
            elif converter == "int":
                schema = {
                    "type": "integer",
                    "format": "int32",
                }
                if "max" in kwargs:
                    schema["maximum"] = kwargs["max"]
                if "min" in kwargs:
                    schema["minimum"] = kwargs["min"]
            elif converter == "float":
                schema = {
                    "type": "number",
                    "format": "float",
                }
            elif converter == "uuid":
                schema = {
                    "type": "string",
                    "format": "uuid",
                }
            elif converter == "path":
                schema = {
                    "type": "string",
                    "format": "path",
                }
            elif converter == "string":
                schema = {
                    "type": "string",
                }
                if "length" in kwargs:
                    schema["length"] = kwargs["length"]
                if "maxlength" in kwargs:
                    schema["maxLength"] = kwargs["maxlength"]
                if "minlength" in kwargs:
                    schema["minLength"] = kwargs["minlength"]
            elif converter == "default":
                schema = {"type": "string"}
            else:
                schema = _parse_custom_url_converter(converter, self.app) or {"type": "string"}

            parameters.append(
                {
                    "name": variable,
                    "in": "path",
                    "required": True,
                    "schema": schema,
                }
            )

        return "".join(subs), parameters

    def request_validation(
        self,
        request: FlaskRequest,
        query: Optional[Type[BaseModel]],
        body: Optional[RequestBase],
        headers: Optional[Type[BaseModel]],
        cookies: Optional[Type[BaseModel]],
    ) -> None:
        raw_query = request.args or None
        if raw_query is not None:
            req_query = parse_multi_dict(raw_query)
        else:
            req_query = {}
        if request.content_type and "application/json" in request.content_type:
            if request.content_encoding and "gzip" in request.content_encoding:
                raw_body = gzip.decompress(request.stream.read()).decode(encoding="utf-8")
                parsed_body = json.loads(raw_body)
            else:
                parsed_body = request.get_json(silent=True) or {}
        elif request.content_type and "multipart/form-data" in request.content_type:
            # It's possible there is a binary json object in the files - iterate through and find it
            parsed_body = {}
            for key, value in request.files.items():
                if value.mimetype == "application/json":
                    parsed_body[key] = json.loads(value.stream.read().decode(encoding="utf-8"))
            # Finally, find any JSON objects in the form and add them to the body
            parsed_body.update(parse_multi_dict(request.form) or {})
        else:
            parsed_body = request.get_data() or {}

        req_headers: Optional[Headers] = request.headers or None
        req_cookies: Optional[Mapping[str, str]] = request.cookies or None
        setattr(
            request,
            "context",
            Context(
                query=query.parse_obj(req_query) if query else None,
                body=(
                    getattr(body, "model").parse_obj(parsed_body)
                    if body and getattr(body, "model")
                    else None
                ),
                headers=headers.parse_obj(req_headers or {}) if headers else None,
                cookies=cookies.parse_obj(req_cookies or {}) if cookies else None,
            ),
        )

    def validate(
        self,
        func: Callable,
        query: Optional[Type[BaseModel]],
        body: Optional[RequestBase],
        headers: Optional[Type[BaseModel]],
        cookies: Optional[Type[BaseModel]],
        resp: Optional[ResponseBase],
        before: Callable,
        after: Callable,
        *args: List[Any],
        **kwargs: Mapping[str, Any],
    ) -> FlaskResponse:
        response, req_validation_error, resp_validation_error = None, None, None
        try:
            self.request_validation(request, query, body, headers, cookies)
        except ValidationError as err:
            req_validation_error = err
            response = make_response(jsonify(err.errors()), self.config.VALIDATION_ERROR_CODE)

        before(request, response, req_validation_error, None)
        if req_validation_error:
            abort(response)  # type: ignore

        response = make_response(func(*args, **kwargs))

        if resp and resp.has_model() and getattr(resp, "validate"):
            model = resp.find_model(response.status_code)
            if model:
                try:
                    model.validate(response.get_json())
                except ValidationError as err:
                    resp_validation_error = err
                    response = make_response(jsonify({"message": "response validation error"}), 500)

        after(request, response, resp_validation_error, None)

        return response

    def register_route(self, app: Flask) -> None:
        self.app = app
        from flask import jsonify

        self.app.add_url_rule(
            self.config.spec_url,
            "openapi",
            lambda: jsonify(self.validator.spec),
        )

        for ui in PAGES:
            self.app.add_url_rule(
                f"/{self.config.PATH}/{ui}",
                f"doc_page_{ui}",
                lambda ui=ui: PAGES[ui].format(self.config),
            )


def _parse_custom_url_converter(converter: str, app: Flask) -> Optional[Dict[str, Any]]:
    """Attempt derive a schema from a custom URL converter."""
    try:
        converter_cls = app.url_map.converters[converter]
        import inspect

        signature = inspect.signature(converter_cls.to_python)
        return_type = signature.return_annotation
        if issubclass(return_type, Enum):
            return {
                "type": "string",
                "enum": [e.value for e in return_type],
            }
    except (KeyError, AttributeError):
        pass
    return None



import logging
from typing import Set, Optional, Dict, Any, Mapping, List


class Config:
    """
    :ivar MODE: mode for route. **normal** includes undecorated routes and
        routes decorated by this instance. **strict** only includes routes
        decorated by this instance. **greedy** includes all the routes.
    :ivar PATH: path for API document page
    :ivar OPENAPI_VERSION: OpenAPI version
    :ivar TITLE: service name
    :ivar VERSION: service version
    :ivar DOMAIN: service host domain
    :ivar VALIDATION_ERROR_CODE: code for validation error responses
    """

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        self.PATH: str = "apidoc"
        self.FILENAME: str = "openapi.json"
        self.OPENAPI_VERSION: str = "3.0.3"
        self.UI: str = "redoc"
        self._SUPPORT_UI: Set[str] = {"redoc", "swagger"}
        self.MODE: str = "normal"
        self._SUPPORT_MODE: Set[str] = {"normal", "strict", "greedy"}
        self.VALIDATION_ERROR_CODE: int = 422

        self.TITLE: str = "Service API Document"
        self.VERSION: str = "0.1"
        self.DOMAIN: Optional[str] = None

        self.INFO: Dict[str, str] = {}
        self.TAGS: List[Mapping[str, str]] = []

        self.logger = logging.getLogger(__name__)

        self.update(**kwargs)

    @property
    def spec_url(self) -> str:
        return f"/{self.PATH}/{self.FILENAME}"

    def __repr__(self) -> str:
        display = "\n{:=^80}\n".format(self.__class__.__name__)
        for k, v in vars(self).items():
            if not k.startswith("__"):
                display += "| {:<30} {}\n".format(k, v)

        return display + "=" * 80

    def update(self, **kwargs: Mapping[str, Any]) -> None:
        """
        update config from key-value pairs

        :param kwargs: key(case insensitive)-value pairs for config

        If the key is not in attributes, it will be ignored. Otherwise, the
        corresponding attribute will be updated. (Logging Level: INFO)
        """
        for key, value in kwargs.items():
            key = key.upper()
            if not hasattr(self, key):
                self.logger.info(f'[✗] Ignore unknown attribute "{key}"')
            else:
                setattr(self, key, value)
                self.logger.info(f'[✓] Attribute "{key}" has been updated to "{value}"')

        assert self.UI in self._SUPPORT_UI, "unsupported UI"
        assert self.MODE in self._SUPPORT_MODE, "unsupported MODE"



PAGES = {
    # https://github.com/Redocly/redoc
    "redoc": """
<!DOCTYPE html>
<html>
    <head>
        <title>{0.TITLE}</title>
        <!-- needed for adaptive design -->
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <link href=
        "https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700"
        rel="stylesheet">

        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
        body {{
            margin: 0;
            padding: 0;
        }}
        </style>
    </head>
    <body>
        <redoc spec-url='{0.spec_url}'></redoc>
        <script src="https://cdn.jsdelivr.net/npm/redoc@next/bundles/redoc.standalone.js"> </script>
    </body>
</html>""",
    # https://swagger.io
    "swagger": """
<!-- HTML for static distribution bundle build -->
<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="UTF-8">
        <title>{0.TITLE}</title>
        <link rel="stylesheet" type="text/css"
        href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css" >
        <style>
        html
        {{
            box-sizing: border-box;
            overflow: -moz-scrollbars-vertical;
            overflow-y: scroll;
        }}

        *,
        *:before,
        *:after
        {{
            box-sizing: inherit;
        }}

        body
        {{
            margin:0;
            background: #fafafa;
        }}
        </style>
    </head>

    <body>
        <div id="swagger-ui"></div>

        <script
src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js"></script>
        <script
src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-standalone-preset.js"></script>
        <script>
        window.onload = function() {{
        // Begin Swagger UI call region
        const ui = SwaggerUIBundle({{
            url: "{0.spec_url}",
            dom_id: '#swagger-ui',
            deepLinking: true,
            presets: [
            SwaggerUIBundle.presets.apis,
            SwaggerUIStandalonePreset
            ],
            plugins: [
            SwaggerUIBundle.plugins.DownloadUrl
            ],
            layout: "StandaloneLayout"
        }})
        // End Swagger UI call region

        window.ui = ui
        }}
    </script>
    </body>
</html>""",
}



from io import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

with open(path.join(here, "requirements/production.txt"), encoding="utf-8") as f:
    requires = [req.strip() for req in f if req]


setup(
    name="flask_pydantic_spec",
    version="0.6.0",
    author="Chris Gearing, Simon Hayward, Rob Young, Donald Fleming, Saurabh Jha",
    author_email="chris.gearing@turntown.digital",
    description=(
        "generate OpenAPI document and validate request & response " "with Python annotations."
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/turner-townsend/flask-pydantic-spec",
    packages=find_packages(exclude=["examples*", "tests*"]),
    package_data={},
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.8",
    install_requires=requires,
    extras_require={
        "flask": ["flask"],
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [],
    },
)
